# Commodore 64 Keyboard Scanner
# @stephenda gmail.com 2020
# https://www.zx81keyboardadventure.com 
# MIT License (https://opensource.org/licenses/MIT) see LICENSE
#
# Software is modified from the original version published
# by MagPi and PJ Evans for his ZX Spectrum keyboard build
#
# Code should work with a Full C64 keyboard, not just the 
# Commodore Pet/Educator 64 Mini keyboard it was writen for.
# If you use this code or improve on it, I'd really enjoy 
# seeing any updates and rewrites you make. (There is much to do)


# ZX Spectrum focused code by mrpjevans
#
# ZX Raspberry Keyboard Scanner v4.1
# @mrpjevans mrpjevans.com 2017
# MIT License (https://opensource.org/licenses/MIT) see LICENSE
#


import time, sys, wiringpi, uinput, os

#
# MagPi Article Mappings
#

# KB1 (BCOM GPIO pins)
dataLines = [26,20,19,16,13,6,12,5]
# KB1 (BCOM GPIO pins)
addressLines = [0,1,23,15,4,14,3,2]


# The Commodore 64 Keyboard Matrix: Mapped to PC keyboard codes
keys = [
	['1',		'3',		'5',		'7',		'9',		'0',		'INSERT',		'BACKSPACE'],
	['GRAVE',	'W',		'R',		'Y',		'I',		'P',		'0',			'ENTER'],
	['TAB',		'A',		'D',		'G',		'J',		'L',		'SEMICOLON',	'RIGHT'],
	['ESC',		'LEFTSHIFT','X',		'V',		'N',		'COMMA',	'SLASH',		'DOWN'],
	['SPACE',	'Z',		'C',		'B',		'M',		'DOT',		'RIGHTSHIFT',	'F1'],
	['LEFTCTRL','S',		'F',		'H',		'K',		'0',		'EQUAL',		'F3'],
	['Q',		'E',		'T',		'U',		'O',		'0',		'DELETE',		'F5'],
	['2',		'4',		'6',		'8',		'0',		'MINUS',	'HOME',			'F7']
]

# The Commodore 64 Keyboard Matrix: Mapped to PC keyboard codes
# Mapping is against a US keyboard layout
# 1st Value: False or True: True if SHIFT is required to map the correct PC Keyboard Character
# 2nd Value: PC Keyboard Key mapped to C64 Matrix
KeysBase = [
	[[False,'1'],[False,'3'],[False,'5'],[False,'7'],[False,'9'],[True,'EQUAL'],[False,'INSERT'],[False,'BACKSPACE']],
	[[False,'GRAVE'],[False,'W'],[False,'R'],[False,'Y'],[False,'I'],[False,'P'],[True,'8'],[False,'ENTER']],
	[[False,'TAB'],[False,'A'],[False,'D'],[False,'G'],[False,'J'],[False,'L'],[False,'SEMICOLON'],[False,'RIGHT']],
	[[False,'ESC'],[False,'LEFTSHIFT'],[False,'X'],[False,'V'],[False,'N'],[False,'COMMA'],[False,'SLASH'],[False,'DOWN']],
	[[False,'SPACE'],[False,'Z'],[False,'C'],[False,'B'],[False,'M'],[False,'DOT'],[False,'RIGHTSHIFT'],[False,'F1']],
	[[False,'LEFTCTRL'],[False,'S'],[False,'F'],[False,'H'],[False,'K'],[True,'SEMICOLON'],[False,'EQUAL'],[False,'F3']],
	[[False,'Q'],[False,'E'],[False,'T'],[False,'U'],[False,'O'],[True,'2'],[False,'DELETE'],[False,'F5']],
	[[False,'2'],[False,'4'],[False,'6'],[False,'8'],[False,'0'],[False,'MINUS'],[False,'HOME'],[False,'F7']]
]


# The Commodore 64 Keyboard Matrix: Mapped to PC keyboard Shifted Key Codes
# Shifted cahacter codes may be different from the KeysBase for some keys to map correctly
# 1st Value: False or True: True if NoSHIFT is required to map the correct PC Keyboard Character
# 2nd Value: PC Keyboard Key mapped to C64 Matrix
KeysShift = [
	[[False,'1'],[False,'3'],[False,'5'],[True,'APOSTROPHE'],[False,'9'],[False,'EQUAL'],[True,'INSERT'],[True,'BACKSPACE']],
	[[False,'GRAVE'],[False,'W'],[False,'R'],[False,'Y'],[False,'I'],[False,'P'],[False,'8'],[False,'ENTER']],
	[[False,'TAB'],[False,'A'],[False,'D'],[False,'G'],[False,'J'],[False,'L'],[True,'RIGHTBRACE'],[True,'LEFT']],
	[[False,'ESC'],[False,'LEFTSHIFT'],[False,'X'],[False,'V'],[False,'N'],[False,'COMMA'],[False,'SLASH'],[True,'UP']],
	[[False,'SPACE'],[False,'Z'],[False,'C'],[False,'B'],[False,'M'],[False,'DOT'],[False,'RIGHTSHIFT'],[True,'F2']],
	[[False,'LEFTCTRL'],[False,'S'],[False,'F'],[False,'H'],[False,'K'],[False,'LEFTBRACE'],[True,'EQUAL'],[True,'F4']],
	[[False,'Q'],[False,'E'],[False,'T'],[False,'U'],[False,'O'],[False,'2'],[False,'DELETE'],[True,'F6']],
	[[False,'APOSTROPHE'],[False,'4'],[False,'7'],[False,'9'],[True,'0'],[True,'MINUS'],[False,'HOME'],[True,'F8']]
]


# The Commodore 64 Keyboard Matrix: - Shifted Key Codes
keysShift = [
	['1',			'3',		'5',		'7',		'0',		'0',		'INSERT',		'BACKSPACE'],
	['GRAVE',		'W',		'R',		'Y',		'I',		'P',		'0',			'ENTER'],
	['TAB',			'A',		'D',		'G',		'J',		'L',		'SEMICOLON',	'LEFT'],
	['ESC',			'LEFTSHIFT','X',		'V',		'N',		'COMMA',	'SLASH',		'UP'],
	['SPACE',		'Z',		'C',		'B',		'M',		'DOT',		'RIGHTSHIFT',	'F1'],
	['LEFTCTRL',	'S',		'F',		'H',		'K',		'0',		'EQUAL',		'F3'],
	['Q',			'E',		'T',		'U',		'O',		'0',		'DELETE',		'F5'],
	['APOSTROPHE',	'4',		'7',		'9',		'0',		'MINUS',	'HOME',			'F7']
]

# The Commodore 64 Keyboard Matrix: - PC mapping require Un-Shiting to mimic shift - or something else


keySets = [
	[
		['1','1','A'], ['3','3','B']
	]
]

# Function key mode
funcKeys = [
	['F1','F2','F3','F4','LEFT','LEFT','LEFT','LEFT'],
	['Q','W','E','R','T','LEFT','LEFT','LEFT'],
	['A','S','D','F','G','LEFT','LEFT','LEFT'],
	['ESC','9','RIGHT','UP','DOWN','LEFT','LEFT','LEFT'],
	['P','O','I','U','Y','LEFT','LEFT','LEFT'],
	['LEFTSHIFT','Z','X','C','V','LEFT','LEFT','LEFT'],
	['ENTER','L','K','J','H','LEFT','LEFT','LEFT'],
	['SPACE','LEFTCTRL','M','N','B','LEFT','LEFT','LEFT']
]

# Track keypresses so we can support multiple keys
keyTrack = [
	[False, False, False, False, False, False, False, False],
	[False, False, False, False, False, False, False, False],
	[False, False, False, False, False, False, False, False],
	[False, False, False, False, False, False, False, False],
	[False, False, False, False, False, False, False, False],
	[False, False, False, False, False, False, False, False],
	[False, False, False, False, False, False, False, False],
	[False, False, False, False, False, False, False, False],
]

# Keyboard mode and reset button
buttonPressed = -1
buttonGPIO = 21
buttonTime = 0

# 0 = Spectrum, 1 = Function Keys
keyboardMode = 0

# Local path
myDir = os.path.dirname(os.path.realpath(__file__));

# Well this is annoying
device = uinput.Device([
        uinput.KEY_A, uinput.KEY_B, uinput.KEY_C, uinput.KEY_D, uinput.KEY_E, uinput.KEY_F, uinput.KEY_G, uinput.KEY_H,
        uinput.KEY_I, uinput.KEY_J, uinput.KEY_K, uinput.KEY_L, uinput.KEY_M, uinput.KEY_N, uinput.KEY_O, uinput.KEY_P,
		uinput.KEY_Q, uinput.KEY_R, uinput.KEY_S, uinput.KEY_T, uinput.KEY_U, uinput.KEY_V, uinput.KEY_W, uinput.KEY_X,
        uinput.KEY_Y, uinput.KEY_Z, uinput.KEY_0, uinput.KEY_1, uinput.KEY_2, uinput.KEY_3, uinput.KEY_4, uinput.KEY_5,
        uinput.KEY_6, uinput.KEY_7, uinput.KEY_8, uinput.KEY_9,
        uinput.KEY_LEFTSHIFT, uinput.KEY_ENTER, uinput.KEY_SPACE, uinput.KEY_LEFTCTRL,
        uinput.KEY_F1, uinput.KEY_F2, uinput.KEY_F3, uinput.KEY_F4, uinput.KEY_F5, uinput.KEY_F6, uinput.KEY_F7, uinput.KEY_F8,
        uinput.KEY_UP, uinput.KEY_DOWN, uinput.KEY_LEFT, uinput.KEY_RIGHT,

        uinput.KEY_ESC, uinput.KEY_BACKSPACE, uinput.KEY_HOME, uinput.KEY_RIGHTSHIFT, uinput.KEY_INSERT, uinput.KEY_RIGHTCTRL,
        uinput.KEY_MINUS, uinput.KEY_COMMA, uinput.KEY_SLASH, uinput.KEY_DELETE, uinput.KEY_EQUAL,
        uinput.KEY_SEMICOLON, uinput.KEY_DOT, uinput.KEY_TAB, uinput.KEY_GRAVE, uinput.KEY_INSERT,
        uinput.KEY_PAGEUP,uinput.KEY_APOSTROPHE,uinput.KEY_RIGHTBRACE,uinput.KEY_LEFTBRACE
        ])

# Setup GPIO
wiringpi.wiringPiSetupGpio()

# Set all address lines high
for addressLine in addressLines:
	wiringpi.pinMode(addressLine, 1)
	wiringpi.digitalWrite(addressLine, 1)

# Set all data lines for input
for dataLine in dataLines:
	wiringpi.pullUpDnControl(dataLine, 2)

# Setup Button
wiringpi.pullUpDnControl(buttonGPIO, 2)

# Announce
print("Running")

try:
	print(keys[3][1])
	print(keys[4][6])
	print(keySets[0][1][2])
	# Loop forever
	while True:

		# Button check
		if(wiringpi.digitalRead(buttonGPIO) == False):

			# Record time the button was pressed
			if(buttonPressed == -1):
				buttonPressed = int(time.time())
				print('Button pressed')

		elif(buttonPressed != -1):

			# Button released - but how long was it pressed for?
			buttonTime = int(time.time()) - buttonPressed

			# If over 3 secs, switch keyboard mode
			if(buttonTime < 5 ):
				if(buttonTime < 2 ):
					# Switch modes
					keyCode = getattr(uinput,'KEY_PAGEUP')
					device.emit(keyCode, 1)
					print("Pressed PAGEUP / Restore")
					time.sleep(0.25)
					device.emit(keyCode, 0)
					print("Releasing PAGEUP / Restore")
				else:
					if(keyboardMode == 0):
						print("Switching to Keys")
						keyboardMode = 1;
					else:
						print("Switching to Keys")
						keyboardMode = 0;

			else:

				print('Shutdown PI')
				os.system('sudo shutdown now')

			# Reset
			buttonPressed = -1

		# Individually set each address line low
		for addressLine in range(8):
			# Set low
			wiringpi.digitalWrite(addressLines[addressLine], 0)

			# Scan data lines
			for dataLine in range(8):
				# Get state and details for this button
				isFree = wiringpi.digitalRead(dataLines[dataLine])
				if(keyboardMode == 0):
					if(keyTrack[3][1] == 1):
						keyPressed = KeysShift[addressLine][dataLine][1]
						if(KeysShift[addressLine][dataLine][0] == True):
							ShiftTrack = False
							device.emit(getattr(uinput, 'KEY_LEFTSHIFT'), 0)
					else:
						keyPressed = KeysBase[addressLine][dataLine][1]
						if(KeysBase[addressLine][dataLine][0] == True):
							ShiftTrack = True
							device.emit(getattr(uinput, 'KEY_LEFTSHIFT'), 1)
				else:
					keyPressed = funcKeys[addressLine][dataLine]
				keyCode = getattr(uinput, 'KEY_' + keyPressed)

				# If pressed for the first time
				if(isFree == False and keyTrack[addressLine][dataLine] == False):

					# Press the key and make a note
					print('Pressing ' + keyPressed)
					print(keyCode)
					device.emit(keyCode, 1)
					keyTrack[addressLine][dataLine] = True

				# If not pressed now but was pressed on last check
				elif(isFree == True and keyTrack[addressLine][dataLine] == True):

					# Release the key and make a note
					print('Releasing ' + keyPressed)
					device.emit(keyCode, 0)
					keyTrack[addressLine][dataLine] = False

			if(keyTrack[3][1] == 1):
				device.emit(getattr(uinput, 'KEY_LEFTSHIFT'),  1)
			elif(keyTrack[3][1] == 0):
				device.emit(getattr(uinput, 'KEY_LEFTSHIFT'),  0)
			# Set high
			wiringpi.digitalWrite(addressLines[addressLine], 1)

			# Have a quick snooze (suggested by Keef)
			time.sleep(.01)


except KeyboardInterrupt:
	sys.exit(0)
